#if defined(OS_WIN32)
#include <windows.h>
#include <malloc.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <fstream>

#include <string.h>
#include <math.h>
#include <ctype.h>
#include <exception>

/// bvqx specific files ///
#include "global.h"
#include "BVQXPluginInterfaceExtern.h"
#include "giftiplugin.h"

// plugin specific files
#include "gifti_io.h"
#include "gifti1file.h"
#include "export.h"
#include "import.h"
#include "info.h"
#include "bvqx_io.h"
#include "constants.h"

/* 
 Versions:
 1.6 (18-10-11): added functions to import CT maps (this was not in stat map range) and copy map name
 1.5 (20-09-11): added functions to export retinotopy maps (except field sign)
 1.4 (02-12-10): added functions to convert POIs and support for gzb64 en b64 data import
 1.3 (05-10-10): added import function for color labeltable
*/

giftiplugin::giftiplugin() {
}

giftiplugin::~giftiplugin() {
}


static const char *pluginDescription = "~~ GIFTI converter (v1.6) ~~<BR>This plugin makes it possible to import and export surface files (meshes, timeseries, statistical maps and patches of interest) to GIFTI format (see <A HREF=\"http://www.nitrc.org/projects/GIFTI\">GIFTI on NITRC</A>).<BR>The manual can be found <b><a href=\"GIFTIconverterHelp/index.html\">here</a></b><br>.";
static const char *pluginName = "GIFTI converter (v1.6)"; 
static const char *authors = "<i>Joost Mulders and Hester Breman</i>";
static const char *helplocation = "GIFTIconverterHelp/index.html";
static const char *guiscriptname = "gifticonverter.js";

enum MainMenuOptions {MAINMENU_IMPORT = 1, MAINMENU_EXPORT, MAINMENU_INFO};

bool giftiplugin::initPlugin() // this function is called in GUI plugins once after class has been created - not called in pure computational plugins
{
	
	initSuccess = false;
	// This function call and check of return value is necessary to check whether access to QX works
	
	int ret = InitQXAccess();
	
	// Never continue, if the "InitQXAccess" function fails
	if(!ret)
		return false;
	
	
	// Test whether the used QX version is new enough to support the commands used in your plugin
	int Major, Minor, SubMinor;
	qxGetVersionOfBrainVoyagerQX(&Major, &Minor, &SubMinor);
	
	// We require v1.4.8
	if( !checkVersionGreaterEqualVersion(Major, Minor, SubMinor, 2, 2, 0) )
	{
		// we can use message box since this was available with the first QX version with plugin support (v1.3.8)
		qxMessageBox("Can not execute this plugin - BrainVoyager QX version 2.2 or higher required.");
		return false;
	}

	initSuccess = true;
	return true;
}

bool giftiplugin::execute() {

	if (!initSuccess) return false;

	int filetypeid;
	char *welcomemsg = new char[256];
	sprintf(welcomemsg, "GIfTI converter %s is running!", pluginversion);
	qxLogText(welcomemsg);

	int mainmenuchoice;

	// let's get some parameters from GUI script:
	char task_name[101];
	qxGetStringParameter("Command", task_name); 
	char InfoString[501];
	sprintf(InfoString, "Command received: %s", task_name);
	qxLogText(InfoString);
	
	// collect acquisition parameters:
	if (!strcmp(task_name, "Import")) {

		filetypeid = qxGetIntParameter("FileTypeId");
		bool doPrint = (bool) qxGetIntParameter("PrintToLogtab");
		bool doShow = (bool) qxGetIntParameter("OpenInBVQX");
		bool doReverse = (bool) qxGetIntParameter("ReverseWinding");
		bool addColorsFile = (bool) qxGetIntParameter("AddColorsFile"); 
		
		// voer uit
		if (!import_gifti(filetypeid, doPrint, doShow, doReverse, addColorsFile)) return false;
	
	} else if (!strcmp(task_name, "Export")) {
		
		filetypeid = qxGetIntParameter("FileTypeId");
		bool doBinaryExport =  (bool) qxGetIntParameter("DoBinaryExport");
		bool useCurrentFile =  (bool) qxGetIntParameter("UseCurrentFile");
		bool doExportColors = (bool) qxGetIntParameter("DoExportColors");

		// voer uit
		if (!export_brainvoyager(filetypeid, useCurrentFile, doBinaryExport, doExportColors)) return false;
				
	} else if (!strcmp(task_name, "Print")) {
		
		filetypeid = qxGetIntParameter("FileTypeId");
		bool doPrintToLog =  (bool) qxGetIntParameter("PrintToLog");	
		bool doPrintToFile =  (bool) qxGetIntParameter("PrintToFile");
		bool useCurrentFile =  (bool) qxGetIntParameter("UseCurrentFile");
		
		// voer uit
		char *filename = new char[512];
		bool exportfromdisk = (useCurrentFile == 0);
		if(!exportfromdisk) sprintf(filename, "none");
		switch (filetypeid){
			case TypeGifti:
				if (!getgiftiinfo((bool)doPrintToFile, (bool)doPrintToLog)) return false;
				break;
			case TypeSrf:
				if (!getsrfinfo(exportfromdisk, (bool)doPrintToFile, (bool)doPrintToLog)) return false;
				break;
			case TypeSmp:
				if (!getsmpinfo(exportfromdisk, (bool)doPrintToFile, (bool)doPrintToLog)) return false;
				break;
			case TypeMtc:
				if (!getmtcinfo(exportfromdisk, (bool)doPrintToFile, (bool)doPrintToLog)) return false;
				break;
			case TypePoi:
				if (!getpoiinfo(exportfromdisk, (bool)doPrintToFile, (bool)doPrintToLog)) return false;
				break;

				
		} // end submenu switch
				
	} else {
		char logtext[256];
		sprintf(logtext, "Received unknown command: %s", task_name);
		qxLogText(logtext);
	
	}

	return true;
}

//
// C access functions (extern "C" in header)
//

// this creates your plugin class
PLUGIN_ACCESS BVQXPluginInterface *createPlugin()
{
	return new giftiplugin();
}

// return a name for your plugin
PLUGIN_ACCESS const char *getPluginName() {
	return pluginName;
}

// returns the descriptive string defined above
PLUGIN_ACCESS const char *getPluginDescription() {

	return pluginDescription;
}

PLUGIN_ACCESS const char *getAuthor()
{
	return authors;
}

// you may provide the name for a HTML file containing online help
//
PLUGIN_ACCESS const char *getHelpFile()
{
	return helplocation;
}

// you may provide a version string ("Major.Minor") for (updated) releases of your plugins
//
PLUGIN_ACCESS const char *getPluginVersion()
{
	return pluginversion;
}

// NEW: If you want to use GUI dialogs, return a GUI script file, if not, return empty C string
//
PLUGIN_ACCESS const char *getGUIScript()
{
	return guiscriptname;
}
