// plugin specific headers
#include <set>
#include "Plugin_SRF_Header.h"
#include "Plugin_MTC_Header.h"
#include "Plugin_SMP_Header.h"
#include "Plugin_POI_Header.h"
#include "BVQXPluginInterfaceExtern.h"
#include "SRF_Header.h"
#include "MTC_Header.h"
#include "SMP_Header.h"
#include "POI_Header.h"
#include "export.h"
#include "bvqx_io.h"
#include "gifti_io.h"
#include "classify.h"
#include "convertbvqxstructs.h"
#include "info.h"
#include "import.h"
#include "constants.h"
#include "surfacefunctions.h"

/* GIFTI_IMAGE STRUCTURE

typedef struct {
// attributes
int               numDA;      // number of DataArrays
char            * version;    // GIFTI version string

// elements
giiMetaData       meta;
giiLabelTable     labeltable;
giiDataArray   ** darray;

// extras
int               swapped;    // were the bytes swapped
int               compressed; // was the data compressed
nvpairs           ex_atrs;    // extra attributes
} gifti_image;

*/

bool import_gifti(int filetypeid, bool doPrint, bool doShow, bool doReverse, bool addColorsFile) {

	bool success = true;
	char *filename = new char [256];
	char *colorsfilename = new char [256];
	int usrresp;
	
	if (filetypeid == TypeSrf) {
		usrresp= qxGetOpenFileName("Please select the GIFTI surface file", "", "*.gii", filename);
		if (addColorsFile && (usrresp > 0)) {
			usrresp= qxGetOpenFileName("Please select the GIFTI colors file", "", "*.gii", colorsfilename);			
			if (usrresp == 0) addColorsFile = false;
		}
	} else if (filetypeid == TypeMtc) {
		usrresp= qxGetOpenFileName("Please select the GIFTI time series file", "", "*.gii", filename);
	} else if (filetypeid == TypeSmp) {
		usrresp= qxGetOpenFileName("Please select the GIFTI statistical file", "", "*.gii", filename);
	} else if (filetypeid == TypePoi) {
		usrresp= qxGetOpenFileName("Please select the GIFTI label file", "", "*.gii", filename);
	}
	
	if (usrresp < 1) {
		qxMessageBox("No GIFTI file was selected. Procedure finished.");
		success =  false;
	}
	if (success) {
		
		// 1. read image
		gifti_image *gii = gifti_read_image(filename, 1);
		headerdump_gifti(gii, false, true, filename); // new per 15-04-09
		gifti_image *giicolors = NULL;
		if (addColorsFile) {
			giicolors = gifti_read_image(colorsfilename, 1);
			headerdump_gifti(giicolors, false, true, colorsfilename);	
		}
		
		// 2. check the filetype and select appropriate import function
		if (gii != NULL) {
			int intent;
			
			std::set<int> intentcontainer;
			for (int i = 0; i<gii->numDA; i++){
				intent = gii->darray[i]->intent;
				if (((intent >= NIFTI_FIRST_STATCODE) && (intent < NIFTI_LAST_STATCODE)) || intent == NIFTI_INTENT_SHAPE)  {
					intent = NIFTI_FIRST_STATCODE;
				}
				intentcontainer.insert(intent);
			}
		
			if (intentcontainer.find(NIFTI_INTENT_POINTSET) != intentcontainer.end()) {
				success = import_srf(gii, filename, doPrint, doShow, doReverse, giicolors);
				if (success) qxLogText("GIFTI surface successfully imported.");
			} else if (intentcontainer.find(NIFTI_INTENT_TIME_SERIES) != intentcontainer.end()) {
				success = import_mtc(gii, filename, doPrint, doShow);
				if (success) qxLogText("GIFTI time series successfully imported.");
			} else if (intentcontainer.find(NIFTI_FIRST_STATCODE) != intentcontainer.end()) {	////!!!		 
				success = import_smp(gii, filename, doPrint, doShow);
				if (success) qxLogText("GIFTI surface map successfully imported.");
			} else if (intentcontainer.find(NIFTI_INTENT_LABEL) != intentcontainer.end()) {
					success = import_poi(gii, filename, doPrint, doShow);
					if (success) qxLogText("GIFTI label table successfully imported.");
			} else {
				qxLogText("GIfTI filetype not supported (check INTENT code by making headerdump).");
				success = false;
			}
		}
	}
	
	return success;
}


bool import_srf(gifti_image *gii, char *filename, bool doPrint, bool doShow, bool doReverse, gifti_image *giicolors){

	bool success = true;

	// 1. declaration of variable(s)
	srfStruct *srf		= new srfStruct;

	// 2. convert to own structure
	char *srfimgname	= exchangeExtension(filename, ".srf");
	srf->origName		= srfimgname;
	srf->reserved		= 0;
	srf->NrOfVertices	= gii->darray[0]->dims[0];
	srf->NrOfTriangles  = gii->darray[1]->dims[0];
	//Get vertex coordinates form gii into our srf structure
	success = verticesGii2Srf(srf, gii);
	srf->MeshCenterX = meanFloatArray(srf->VertexX, srf->NrOfVertices);
	srf->MeshCenterY = meanFloatArray(srf->VertexY, srf->NrOfVertices);
	srf->MeshCenterZ = meanFloatArray(srf->VertexZ, srf->NrOfVertices);
	srf->ConvCurvAlpha = 1;
	srf->ConvCurvBlue = 0.980;
	srf->ConvCurvGreen = 0.733;
	srf->ConvCurvRed = 0.332;
	srf->ConcCurvAlpha = 1;
	srf->ConcCurvBlue = 0.320;
	srf->ConcCurvGreen = 0.240;
	srf->ConcCurvRed = 0.1;
	srf->fileVersion = 3;
	
	// ~~~ color conversion ~~~
	//giiLabelTable *colortable = new giiLabelTable;
	// check whether color labeltable is present
	gifti_image *color_image = gii;
	// check whether external color labeltable is present 
	if (!giicolors) {
		qxLogText("No external color file.");
	} else {
		color_image = giicolors;
	}
	// convert colors (check existence in function)
	char *tmp = new char[256];
	sprintf(tmp, "Length color labeltable: %i, number of vertices: %i", color_image->labeltable.length, srf->NrOfVertices);
	qxLogText(tmp);
	int *meshcolor = convertLabelTableToMeshColors(color_image, srf);
	srf->MeshColor = meshcolor;

	success = trianglesGii2Srf(srf, gii);
	if (doReverse)	success = rewindTriangles(srf);

	srf->UseTriangleStrip = 0;
	srf->nrOfTriangleStripElements = 0;
	srf->flag_bvqxhdr = false;

	success = computeVertexNormals(srf);

	//Colors hopefully set by default using qxCreateSrf
	// 1. Zet de kleuren van labeltable in eigen struct int *meshcolor
	// 2. in convertbvqxstructs.cpp, zet de kleuren over
	// 3. kijk of alles goed opgeslagen wordt (doet BrainVoyager)

	/* 1. tellen van de vertex-paren/triples van driehoeken die bij een vertex betrokken zijn (voor aanmaken neighbour-array (int). */
	// functieaanroep
	int *nrNeighbours = new int[srf->NrOfVertices];
	success = countNeighbours(srf, nrNeighbours);

	/*2. verzamelen van de vertex-paren/triples van driehoeken die bij een vertex betrokken zijn.
	Hier kunnen dus vertices meerdere keren voorkomen.*/
	// functieaanroep>>
	int maxNeighbours = 0;
	for(int i = 0; i<srf->NrOfVertices; i++) if(maxNeighbours<nrNeighbours[i]) maxNeighbours = nrNeighbours[i];
	int** doubleNeighboursList = create2dIntArray(srf->NrOfVertices, maxNeighbours*2);
	success = collectNeighbours(srf, doubleNeighboursList);


	/*3. ordenen van de neighbours zodat een juiste winding ontstaat.
	Dubbele vertices verwijderen.
	*/
	// >> functie()
	neighbours *neighb = new neighbours[srf->NrOfVertices];
	success = eliminateNeighbours(srf, neighb, doubleNeighboursList, nrNeighbours);
	srf->neighb = neighb;
	delete2dIntArray(doubleNeighboursList, srf->NrOfVertices, maxNeighbours*2);

	// 3. save
	success = writesrf(srf);

	if (doPrint) headerdump_srf(srf, false, true); // new per 15-04-09

	char *txt = new char[512];
	sprintf(txt, "SRF saved as:\n%s", srf->origName);
	qxLogText(txt);

	// 4. Delete structs and arrays?

	
	// 5. show (if doShow)
	// voer bestandsnaam terug aan script
	if (doShow) qxSetStringParameter("SrfFileName", srf->origName);
//	if (addColorsFile) qxSetStringParameter("ColorsFileName", ColorsFileName); // NOG BESLISSEN WAT TE DOEN
	
	return success;

}

bool import_mtc(gifti_image *gii, char *filename, bool doPrint, bool doShow){

	bool success = true;
	// 1. declaration of variable(s)
	mtcStruct *mtc		= new mtcStruct;

	// 2. convert to own structure (?)
	char *mtcimgname	= exchangeExtension(filename, ".mtc");
	mtc->origName		= mtcimgname;
	mtc->nrOfVertices	= gii->darray[0]->dims[0]; // even checken of the vertices in de eerste dim zitten.
	mtc->nrOfVolumes	= gii->numDA;
	char *vtcFileName	= "<none>";
	mtc->vtcFileName	= vtcFileName;
	char *prtFileName	= "<none>";
	mtc->prtFileName	= prtFileName;
	mtc->hemoDynDelay	= 0.0;
	mtc->TR				= 0.0;
	mtc->HRFdelta		= 0.0;
	mtc->HRFtau			= 0.0;
	mtc->segmentSize	= 0; // intervals per stimulus block/event) (*1)
	mtc->segmentOffset	= 0;
	mtc->MTCDataType	= 1; // let op! casten naar float if not float // let op 2: char
	mtc->flag_bvqxhdr	= false;

	// ~~~~~ data, loop over volumes ~~~~~
	int volnr;
	float *data = new float[mtc->nrOfVertices*mtc->nrOfVolumes];
	int nrvertices = mtc->nrOfVertices;
	int vertexnr;
	float *timepoint;
	for (volnr=0; volnr < mtc->nrOfVolumes; volnr++) {
		timepoint = static_cast<float*>(gii->darray[volnr]->data);
		for (vertexnr =0; vertexnr < nrvertices; vertexnr++) {
			data[(volnr*nrvertices + vertexnr)] = timepoint[vertexnr];
		}
	} // end of loop over timepoints
	mtc->data = data;


	// 3. create bvqx structure
	success = convertMTC(mtc);
	if (doPrint) headerdump_mtc(mtc, false, true);
	
	// 4. save
	qxSaveMTCOfCurrentSRF(mtc->origName); //////////////// CHECK!!!! schrijf met eigen functie ///////
	char *txt = new char[512];
	sprintf(txt, "MTC saved as:\n%s", mtc->origName);
	qxLogText(txt);
	
	// 5. show (if doShow)
	if (doShow) {
		int loadSuccess = qxLinkMTCToCurrentSRF(mtc->origName, 0);
		if (loadSuccess == 0) qxMessageBox("Mesh time course file (MTC) could not be loaded on current surface file (SRF).");
	}
	return success;
}

bool import_smp(gifti_image *gii, char *filename, bool doPrint, bool doShow){

	bool success = true;
	// 1. create own structure
	smpStruct *smp = new smpStruct;

	// 2. convert to own structure
	char *smpimgname	= exchangeExtension(filename, ".smp");
	smp->origName		= smpimgname;
	smp->nrOfVertices	= gii->darray[0]->dims[0]; // even checken of the vertices in de eerste dim zitten.
	smp->nrOfMaps		= gii->numDA;
	smp->versionNr		= 5; // added 16-12-10
	char *srfname = new char[256];
	sprintf(srfname, "%s", "untitled.srf");
	smp->srfFileName = srfname; // added 16-12-10

	int mapnr;
	srfStatMap *maps = new srfStatMap[smp->nrOfMaps]; // added 16-12-10
	smp->srfStatMaps = maps;							// added 16-12-10
	for (mapnr = 0; mapnr < smp->nrOfMaps; mapnr++) {
		smp->srfStatMaps[mapnr].mapType = getBVQXmaptype(gii->darray[mapnr]->intent);
		if (smp->srfStatMaps[mapnr].mapType == BVQX_MAPTYPE_UNKNOWN){
			qxLogText("!!!!!!!\nWarning: Maptype unknown in BrainVoyager, it is set to T statistic\n!!!!!!!");
			smp->srfStatMaps[mapnr].mapType = BVQX_MAPTYPE_T_STAT;
		}
		// = new char[256];
		char *mapname = getBVMapName(gii, mapnr);
		smp->srfStatMaps[mapnr].surfaceMapName = mapname;

		// below all added 16-12-10 (mostly for headerdump to work)
		
		smp->srfStatMaps[mapnr].mapData = (float*)gii->darray[mapnr]->data; /////////////
		smp->srfStatMaps[mapnr].clusterSize = 25;
		smp->srfStatMaps[mapnr].enableClusterCheck = 1;
		smp->srfStatMaps[mapnr].critValThresh = 2.282; // lower threshold
		smp->srfStatMaps[mapnr].maxThresh = 8.0; // upper threshold
		smp->srfStatMaps[mapnr].useValuesAboveThreshold = 1; // added 07-01-11
		smp->srfStatMaps[mapnr].enableSMPcolor = 0;
		smp->srfStatMaps[mapnr].bonfCorr = 0;
		smp->srfStatMaps[mapnr].RGBLowerThreshPos[0] = 192; // set arbitrary values
		smp->srfStatMaps[mapnr].RGBLowerThreshPos[1] = 0;
		smp->srfStatMaps[mapnr].RGBLowerThreshPos[2] = 0;
		smp->srfStatMaps[mapnr].RGBUpperThreshPos[0] = 255;
		smp->srfStatMaps[mapnr].RGBUpperThreshPos[1] = 192;
		smp->srfStatMaps[mapnr].RGBUpperThreshPos[2] = 0;
		smp->srfStatMaps[mapnr].RGBLowerThreshNeg[0] = 0;
		smp->srfStatMaps[mapnr].RGBLowerThreshNeg[1] = 0;
		smp->srfStatMaps[mapnr].RGBLowerThreshNeg[2] = 128;	
		smp->srfStatMaps[mapnr].RGBUpperThreshNeg[0] = 0;
		smp->srfStatMaps[mapnr].RGBUpperThreshNeg[1] = 128;
		smp->srfStatMaps[mapnr].RGBUpperThreshNeg[2] = 255;			
		smp->srfStatMaps[mapnr].transparentColorFactor = 1.0;
		smp->srfStatMaps[mapnr].showPosNegFlag = 3; // 1: pos 2: neg 3: both (from vmp, added 07-01-11)
		
		// invoke function to obtain distribution parameters
		float *distrparams = new float[3]; // in NIfTI & GIFTI there are always 3 possible intent params, in BrainVoyager 2 (DF=DOF).
		if (!getdistributionparams(gii, mapnr, smp->srfStatMaps[mapnr].mapType, distrparams)) {
			qxLogText("Distribution parameters missing. Set to zero!");
			return false;
		}

		smp->srfStatMaps[mapnr].df1 = distrparams[0];
		smp->srfStatMaps[mapnr].df2 = distrparams[1];

		// create LUT name
		char *lutname = new char[256];
		sprintf(lutname, "<default>");
		smp->srfStatMaps[mapnr].LUTName = lutname;
		
	}

	// 3. create bvqx structure
	smp->flag_bvqxhdr	= false;
//	success = convertSMP(smp); // relocated to doShow condition below

	// 4. save
//	qxSaveSMPsOfCurrentSRF(smp->origName);
	success = writesmp(smp);
	char *txt = new char[512];
	sprintf(txt, "SMP saved as:\n%s", smp->origName);
	qxLogText(txt);
	if (doPrint) headerdump_smp(smp, false, true);
	
	// 5. show (if doShow)
	if (doShow) {
		success = convertSMP(smp); 
		int loadSuccess = qxLoadSMPsToCurrentSRF(smp->origName);
		if (loadSuccess == 0) qxMessageBox("Surface map (SMP) could not be loaded on current surface (SRF).");
	}	
	return success;
}

bool import_poi(gifti_image *gii, char *filename, bool doPrint, bool doShow){

	bool success = true;
	char *txt = new char[512];	
	int sparsity = -1;
	int labeltable = -1;
	// 1. create own structure
	poiStruct *poi = new poiStruct;
	
	// 2. convert to own structure
	char *poiimgname	= exchangeExtension(filename, ".poi");
	poi->POIsFileName		= poiimgname;
	
	// search for signs of sparsity:
	for (int i=0; i<gii->numDA; i++) {
		if (gii->darray[i]->intent == NIFTI_INTENT_NODE_INDEX) sparsity = i;
		if (gii->darray[i]->intent == NIFTI_INTENT_LABEL) labeltable = i;
	}
	
	if (labeltable == -1){
		qxMessageBox("No label data found in GIFTI file!\nCannot convert to POI.");
		return false;
	}
	
	/* 16-12-10: this is a fix for loading POI (does not work when nr of vertices in POI file is not equal to 
	 the number of vertices in the surface, but this number is not recorded in the GIFTI label file when it is sparse */
	qxLogText("Determine number of mesh vertices...");
	if (sparsity > -1) {
		// read nr of vertices from surf
		char *filename = new char[501];
		SRF_Header *bvqxsrfhdr = new SRF_Header;
        bool getcurrent = (bool)qxGetCurrentSRF(bvqxsrfhdr);
        if (getcurrent) { 
			poi->NrOfMeshVertices = bvqxsrfhdr->NrOfVertices;
			sprintf(txt, "Number of mesh vertices: %i. Please note that it is assumed that the imported POI belongs to the currently opened surface file.\nIf the files do not belong together, the location of the POI be incorrect.", poi->NrOfMeshVertices);
			qxLogText(txt);			
		} else {
			int usrresp = qxGetOpenFileName("Please select the BrainVoyager surface file where this POI belongs to", "", "*.srf", filename);
			if (usrresp > 0) { 
				srfStruct *srf = getsrf(filename, (!getcurrent));
				if (srf->fileVersion != -1.0f) {
					poi->NrOfMeshVertices = srf->NrOfVertices;
					sprintf(txt, "Number of mesh vertices: %i", poi->NrOfMeshVertices);
					qxLogText(txt);
				} else {
					poi->NrOfMeshVertices = gii->darray[labeltable]->dims[0];
					qxLogText("Number of mesh vertices not determined.");
				}
			} else {
				poi->NrOfMeshVertices = gii->darray[labeltable]->dims[0];
				qxLogText("Number of mesh vertices not determined.");				
			}
		}
	} else {
		poi->NrOfMeshVertices = gii->darray[labeltable]->dims[0];
		sprintf(txt, "Number of mesh vertices equal to number of POI nodes: %i", poi->NrOfMeshVertices);
		qxLogText(txt);
	} 
		
	poi->FileVersion = 2;
	poi->FromMeshFile = "";
	poi->NrOfPOIs = gii->labeltable.length; // number of pois = number of labels
	poi->NrOfPOIMTCs = 0;
	
	//allocate space for sparse indices with maximum possible size (= NrOfVertices)
	int *node_index;
	if (sparsity != -1)	node_index = static_cast<int*>(gii->darray[sparsity]->data);
	
	// loop over labels: search for each color through datablock
	int rgbval, key;
	int *labelblock;
	poiSubStruct *patch = new poiSubStruct[poi->NrOfPOIs];
	for (int i=0; i<poi->NrOfPOIs; i++){
		
		char *POIName = new char[256];
		sprintf(POIName, gii->labeltable.label[i]);
		patch[i].POIName = POIName;
		int red			= (gii->labeltable.rgba[i*4]*255+.5);
		int green		= (gii->labeltable.rgba[i*4+1]*255+.5);
		int blue		= (gii->labeltable.rgba[i*4+2]*255+.5);
		float alpha		= gii->labeltable.rgba[i*4+3];
		int startval	= 0;//1056964608;
		if (alpha == 0.0f) {
			rgbval = startval;
		} else {
			rgbval = startval + red * pow(16.0f,4) + green * pow(16.0f,2) + blue;
		}
		patch[i].ColorOfPOI = rgbval; //Display color, from RGB + startvalue;		
		
		key = gii->labeltable.key[i];
		int patchsize = 0;
		labelblock = (int*)gii->darray[labeltable]->data; 
		int sizelabeltable = gii->darray[labeltable]->dims[0];
		for (int j = 0; j < sizelabeltable; j++) { // 16-12: was: j < NrOfMeshVertices
			if (labelblock[j] == key) patchsize ++;
		}
		patch[i].NrOfVertices = patchsize;  
		int *POIdata = new int[patchsize]; // little block just for this POI; 16-12: was: NrOfMeshVertices
		int counter = 0;
		for (int j = 0; j < sizelabeltable; j++) { // 16-12: was: NrOfMeshVertices
			if (labelblock[j] == key) {
				if (sparsity == -1){
					POIdata[counter] = j;
				}else{
					POIdata[counter] = node_index[j];
				}
				counter++;
			}
		}	
		patch[i].ReferenceVertex = POIdata[0];    // New in BVQX v2.0: LabelVertex: ... >>>> ask RG
		patch[i].InfoFileName = "";  // New in BVQX v2.0	InfoTextFile: ... >>> ask RG in which cases this is created	
		
		// attach data to substruct
		patch[i].POIdata = POIdata;
	}
	// attach 
	poi->poi = patch;
	 
	 // BrainVoyager specific header
	 bool flag_bvqxhdr;
	
	// 3. don't create bvqx structure
	poi->flag_bvqxhdr	= false;
	
	// 4. save
	writepoi(poi);
	sprintf(txt, "POI saved as:\n%s", poi->POIsFileName);
	qxLogText(txt);
	if (doPrint) headerdump_poi(poi, false, true);
	
	// 5. show (if doShow)
	if (doShow) {
		int loadSuccess = qxLoadPOIs(poi->POIsFileName);
		if (loadSuccess == 0) qxMessageBox("Patch of interest (POI) could not be loaded on current surface (SRF).");
	}	
	return success;
}
